import numpy as np 
#from lightcurves.LC import LightCurve
from lightcurves.HOP import Hopject

import logging
logging.basicConfig(level=logging.ERROR)
"""
set logging to the desired level
logging options:
DEBUG:      whatever happens will be thrown at you
INFO:       confirmation that things are working as expected
WARNING:    sth unexpected happened
ERROR:      sth didn't work, abort mission
""" 

class HopFinder():
    """
    This is an abstract class that resembles an interface. i.e.
    methods that don't do anything but have to be overwritten with 
    children (inheriting classes):
        - HopFinderBaseline
        - HopFinderProcedure
    
    An object inherited from HopFinder can be used to characterize flares, i.e.
    determine start, peak and end time, based on Bayesian blocks with different methods:
        1. baseline:
            Original method as described in Meyer et al. 2019
            https://ui.adsabs.harvard.edu/abs/2019ApJ...877...39M/abstract 
        2. half:
            Start/end is at center of valley block
        3. sharp:
            Neglect valley block
        4. flip:
            Extrapolate flare behavior

    lc_edges:
        a) 'neglect'
            single start and end times are neglected
            incomplete flares (peaks without start or end time) are conservatively neglected
        b) 'add'
            single start and end times are neglected
            if peak has no start/end it is artificially added in beginning/end of light curve

    returns: list of Hopjects, see HOP.py
    """
    def __init__(self, lc_edges='neglect'):
        self.lc_edges = lc_edges

    def find_start_end(self, lc):
        raise NotImplementedError

    def find_peaks(self, lc):
        raise NotImplementedError

    def find(self, lc): 
        starts, ends = self.find_start_end(lc)
        peaks = self.find_peaks(lc)
        peaks, starts, ends = self.clean(peaks, starts, ends, lc)
        if peaks is None:
            logging.info('no hop in this light curve')
            return None 
        peaks, starts, ends = self.clean_multi_peaks(peaks, starts, ends, lc)
        if peaks is None:
            logging.info('no hop in this light curve')
            return None 
        hops = []
        for p, s, e in zip(peaks, starts, ends):
            #TBD hier könnte man noch Kriterien für hopject einfügen (e.g. bins per block/hop)
            hops.append(Hopject((s,p,e), lc, method=type(self).__name__ ))
            ## type(self).__name__ = Name der Klasse
        return hops

    def clean(self, peaks, starts, ends, lc):
        """
        Handle mismatches and issues with peak_time, start_time, and end_time combinations
        lc_edges:
            a) neglect:
                single start and end times are neglected
                peaks without start or end time are neglected
            b) add:
                single start and end times are neglected
                peaks without start/end: start/end is added in beginning/end of light curve
        """
        if len(peaks) < 1:
            logging.info('not variable enough, no peak found')
            return(None, None, None) 
        if self.lc_edges == 'neglect':
            if len(starts) < 1 or len(ends) < 1:
                logging.info('not variable enough, missing start or end')
                return(None, None, None)
        if self.lc_edges == 'add':
            if len(starts) < 1:
                starts = np.insert(starts, 0, lc.edges[0])
                logging.info('inserted single start in beginning of LC')
            if len(ends) < 1:
                ends = np.append(ends,lc.edges[-1])
                logging.info('inserted single end in end of LC')
        if ends[0] < peaks[0]:
            ends = np.delete(ends, 0)
            logging.info('deleted single end in beginning of LC')
            if len(ends) < 1 and self.lc_edges == 'neglect':
                logging.info('this was the only end, not variable enough')
                return(None, None, None)
            if len(ends) < 1 and self.lc_edges == 'add':
                ends = np.append(ends, lc.edges[-1])
                logging.info('inserted single end in end of LC and this is the only end')    
        if starts[-1] > peaks[-1]:
            starts = np.delete(starts, -1)
            logging.info('deleted single start in end of LC')
            if len(starts) < 1 and self.lc_edges == 'neglect':
                logging.info('this was the only start, not variable enough')
                return(None, None, None)
            if len(starts) < 1 and self.lc_edges == 'add':
                starts = np.insert(starts, 0, lc.edges[0])
                logging.info('inserted single start in beginning of LC; this is the only start')
        if peaks[0] < starts[0]:
            if self.lc_edges == 'add':
                # artificially add start
                starts = np.insert(starts, 0, lc.edges[0])
                logging.info('inserted single start in beginning of LC')
            if self.lc_edges == 'neglect':
                # conservatively dismiss first peak if there are multiple peaks
                while ends[0] > peaks[1]:
                    peaks = np.delete(peaks, 0)
                    logging.info('neglected first multiple peak in beginning of LC')
                #conservatively dismiss first peak and first end
                peaks = np.delete(peaks, 0)
                ends = np.delete(ends, 0)
                logging.info('start missing, neglected peak and end in beginning of LC')
                if len(peaks) < 1 or len(ends) < 1:
                    logging.info('this was the only peak or end, not variable enough')
                    return(None, None, None)
        if peaks[-1] > ends[-1]:
            if self.lc_edges == 'add':
                # artificially add end
                ends = np.append(ends, lc.edges[-1])
                logging.info('inserted single end in end of LC') 
            if self.lc_edges == 'neglect':
                # conservatively dismiss last peak if there are multiple peaks
                if len(peaks) > 2:
                    while starts[-1] < peaks[-2]:
                        peaks = np.delete(peaks, -1)
                        logging.info('neglected last multiple peak in end of LC')
                    # conservatively dismiss last peak and last start
                peaks = np.delete(peaks, -1)
                starts = np.delete(starts, -1)
                logging.info('neglected peak and start in end of LC')
                if len(peaks) < 1 or len(starts) < 1:
                    logging.info('this was the only peak or start, not variable enough')
                    return(None, None, None)

        return(peaks, starts, ends)

    def clean_multi_peaks(self, peaks, starts, ends, lc):
        # baseline method could result in multiple peaks within one HOP 
        # -> neglect smaller peak (not so senseful..)
        while len(ends) < len(peaks):
            for x,_ in enumerate(ends):
                if ends[x] > peaks[x+1]:
                    if (lc.block_val[lc.bb_i(peaks[x])] 
                        < lc.block_val[lc.bb_i(peaks[x+1])]):
                            peaks = np.delete(peaks, x)
                    elif (lc.block_val[lc.bb_i(peaks[x])] 
                          >= lc.block_val[lc.bb_i(peaks[x+1])]):
                            peaks = np.delete(peaks, x+1)
                    logging.info('neglected double peak in HOP ' + str(x))
                    break
        return peaks, starts, ends


#----------------------------------------------------------------------------------------------
class HopFinderBaseline(HopFinder):
    """
    BASELINE METHOD
        see Meyer et al. 2019 https://ui.adsabs.harvard.edu/abs/2019ApJ...877...39M/abstract
        Determine peak_time of flare to be at center of colal maxima of the blocks
        Determine start_time/end_time to be where flux exceeds/goes under baseline

    lc.baseline: 
        e.g. mean of flux (default), median of flux, quiescent background ...
    """
    def find_peaks(self, lc):
        diff = np.diff(lc.block_val)
        peaks = [] #time of all local peaks over baseline (in units of edges = units of time)
        for i in range(1,len(diff)):
            # if previous rising; this falling
            if diff[i-1] > 0 and diff[i] < 0:
                if lc.block_val[i] > lc.baseline:
                    # peak_time = middle of peak block
                    peaks.append(lc.edges[i] + (lc.edges[i+1] - lc.edges[i]) /2)
        return peaks

    def find_start_end(self, lc):
        starts = []  
        ends = []    
        for i in range(len(lc.block_val)-1):
            # if this smaller; next one higher
            if lc.block_val[i] < lc.baseline and lc.block_val[i+1] > lc.baseline:
                starts.append(lc.edges[i+1])
            # if this larger; next one lower
            if lc.block_val[i] > lc.baseline and lc.block_val[i+1] < lc.baseline:
                ends.append(lc.edges[i+1])
        return starts, ends

#----------------------------------------------------------------------------------------------
class HopFinderProcedure(HopFinder):
    """
    This is another abstract class that resembles an interface. i.e.
    methods that don't do anything but have to be overwritten with 
    children (inheriting classes):
            - HopFinderHalf
            - HopFinderFlip
            - HopFinderSharp

    Determine peak_time of flare to be at center of colal maxima of the blocks
    Use self.change_point() to determine start and end_time depending on method             
    """
    def find_peaks(self, lc):
        diff = np.diff(lc.block_val)
        peaks = [] # time of all local peaks (units of edges, i.e. units of time)
        for i in range(1,len(diff)):
            # peak = previous rising; this falling
            if diff[i-1] > 0 and diff[i] < 0:
                # peak_time = middle of peak block
                peaks.append(lc.edges[i] + (lc.edges[i+1] - lc.edges[i]) /2)
        return peaks

    def change_point(self, edges, i):
        raise NotImplementedError

    def find_start_end(self, lc):
        starts = []
        ends = []
        diff = np.diff(lc.block_val)
        for i in range(1,len(diff)):
            # change = previous falling; this rising
            if diff[i-1] < 0 and diff[i] > 0: 
                start, end = self.change_point(lc.edges, i)
                starts.append(start)
                ends.append(end)
        return starts, ends

#----------------------------------------------------------------------------------------------
class HopFinderHalf(HopFinderProcedure):
    """
    Determine start/end of flare to be at center of valley block
    """
    def change_point(self, edges, i):
        half_block_time = (edges[i+1] - edges[i]) / 2
        return edges[i+1] - half_block_time, edges[i] + half_block_time

class HopFinderFlip(HopFinderProcedure):
    """
    Extrapolate behavior of flare by flipping adjacent block onto valley block
    Note: half method is used to avoid overlap (i.e. when flip > 1/2 valley block)
    """
    def change_point(self, edges, i):
        half_block_time = (edges[i+1] - edges[i]) / 2
        #clap previous block onto change block
        clap_from_left = edges[i] - edges[i-1]
        #clap following block onto change block
        clap_from_right = edges[i+2] - edges[i+1]
        e = edges[i] + min(half_block_time, clap_from_left)
        s = edges[i+1] - min(half_block_time, clap_from_right)
        return s,e

class HopFinderSharp(HopFinderProcedure):
    """
    Neglect valley block
    """
    def change_point(self, edges, i):
        return edges[i+1], edges[i]





import numpy as np 
from matplotlib import pyplot as plt
import astropy.stats.bayesian_blocks as bblocks
#https://docs.astropy.org/en/stable/api/astropy.stats.bayesian_blocks.html
from lightcurves.HopFinder import *

import logging
logging.basicConfig(level=logging.ERROR)
"""
set logging to the desired level
logging options:
DEBUG:      whatever happens will be thrown at you
INFO:       confirmation that things are working as expected
WARNING:    sth unexpected happened
ERROR:      sth didn't work, abort mission
""" 


def flux_puffer(flux, threshold, threshold_error):
    """
    ATTENTION! This returns artificial flux values! Use cautiously if at all..
    Set every flux bin under threshold to threshold before initializing light curve
    Apply Bayesian blocks -> detect significant variations wrt threshold = flares?
    """
    flux_new = np.where(flux > threshold, flux, threshold)
    flux_error_new = np.where(flux > threshold, flux_error, th_error)
    return(flux_new, flux_error_new)

def fix_data(time, flux, flux_error):
    """
    ATTENTION! this deletes bins, if there is np.nan in flux(_error) or duplicate in time
    """
    flux_ = flux[np.invert(np.isnan(flux)) * np.invert(np.isnan(flux_error))]
    flux_error_ = flux_error[np.invert(np.isnan(flux)) * np.invert(np.isnan(flux_error))]
    time_ = time[np.invert(np.isnan(flux)) * np.invert(np.isnan(flux_error))]
    logging.info('Deleted ' + str(len(flux) - len(flux_)) + ' np.nan values.')
    unique_time, unique_time_id = np.unique(time_, return_index=True)
    good_flux = flux_[unique_time_id]
    good_flux_error = flux_error_[unique_time_id]
    logging.info('Deleted ' + str(len(time_) - len(unique_time)) + ' time duplicates')
    return(unique_time, good_flux, good_flux_error)


def get_gti_iis(time, n_gaps, n_pick):
    # get index of good time intervals (divide LC into secitons in case there are n_gaps gaps in data; like FACT)
    # biggest time gaps (length in s) in chronological order
    diff = np.array([t - s for s, t in zip(time, time[1:])])
    diff1 = np.sort(diff)
    ii = [x for x in range(len(diff)) if diff[x] in diff1[-n_gaps:]] #index of the 10 longest gaps
    GTI_start_ii = np.array(ii)+1
    GTI_start_ii = np.insert(GTI_start_ii,0,0)
    GTI_end_ii = np.array(ii)
    GTI_end_ii = np.append(GTI_end_ii, len(time)-1)
    if n_pick:
        # only consider the n_pick longest gtis 
        gap_len = np.array([t - s for s,t in zip(GTI_start_ii, GTI_end_ii)])
        gap_len1 = np.sort(gap_len)
        ii = [x for x in range(len(gap_len)) if gap_len[x] in gap_len1[-n_pick:]] # n_gaps = considered gaps (longest not gaps)
        GTI_start_ii_ = GTI_start_ii[ii]
        GTI_end_ii_ = GTI_end_ii[ii]
        return GTI_start_ii_, GTI_end_ii_
    else:
        return GTI_start_ii, GTI_end_ii

def make_gti_lcs(lc, n_gaps, n_pick=None):
    """
    Divide one lc with n_gaps gaps into several lcs with good coverage.
    """
    gti_starts, gti_ends = get_gti_iis(lc.time, n_gaps, n_pick)
    if n_pick is None:
        n_pick = n_gaps + 1 #select all 
    chunks = []
    for g in range(n_pick):
        gti_lc = LightCurve(lc.time[gti_starts[g]:gti_ends[g]+1], 
                            lc.flux[gti_starts[g]:gti_ends[g]+1], 
                            lc.flux_error[gti_starts[g]:gti_ends[g]+1],
                            name=lc.name, z=lc.z)
        chunks.append(gti_lc)
    return(np.array(chunks))


#--------------------------------------------------------------------------------------------------
class LightCurve:
    """
    Light Curve Class
    ------------------
    Create a light curve based on input data: time, flux, flux_error
    """
    def __init__(self, time, flux, flux_error, name=None, z=None, telescope=None):
        self.time = np.array(time)
        self.flux = np.array(flux)
        self.flux_error = np.array(flux_error)
        self.name = name
        self.z = z
        self.telescope = telescope
        if len(time) != len(flux) or len(time) != len(flux_error):
            raise ValueError('Input arrays do not have same length')
        if len(flux[np.isnan(flux)]) > 0 or len(flux_error[np.isnan(flux_error)]) > 0:
            raise TypeError('flux or flux_error contain np.nan values')
        if len(time) != len(np.unique(time)):
            raise ValueError('time contains duplicate values')

    def plot_lc(self, data_color='k', ax=None, **kwargs):
        if ax is None:
            ax = plt.gca()
        ax.errorbar(x=self.time, y=self.flux, yerr=self.flux_error, ecolor=data_color, 
                     elinewidth=1, linewidth=0, marker='+', markersize=3, 
                     color=data_color, **kwargs)

    def plot_hline(self, value, ax=None, **kwargs):
        if ax is None:
            ax = plt.gca()
        ax.hlines(value, xmin=min(self.time), xmax=max(self.time), **kwargs)

    #----------------------------------------------------------------------------------------------
    def get_bblocks(self, gamma_value=None, p0_value=0.05): 
        """
        Bayesian block algorithm (https://ui.adsabs.harvard.edu/abs/2013arXiv1304.2818S/abstract)
        fitness is set to 'measures' since we assume Gaussian error for flux measurements
        from astropy (https://docs.astropy.org/en/stable/api/astropy.stats.bayesian_blocks.html)
        Returns edges of blocks (significant changes of flux) in units of time (e.g. MJD)
        Edges are converted to edge_index (based on position in time array)
        -> See GitHub description and Jupyter Notebook for more information
        block_val are the flux values of all blocks based on the mean of all flux bins within
        block_val_error is the corresponding error computed with Gaussian error propagation
        block_pbin has the same shape as flux and is filled with corrsponding block values
        """
        # get Bayesian block edges for light curve
        self.edges = bblocks(t=self.time, x=self.flux, sigma=self.flux_error, fitness='measures',
                             gamma=gamma_value, p0=p0_value)
        logging.debug('got edges for light curve')

        if len(self.edges) <= 2:
            logging.warning('light curve is constant; only one bayesian block found.')
            self.block_pbin = np.ones(len(self.flux)) * np.mean(self.flux)
            self.block_val = np.array([np.mean(self.flux)])
            self.block_val_error = np.array([np.std(self.flux)])
            self.edge_index = np.array([0, -1])
            self.edges = np.array([self.time[0], self.time[-1]])
            return(self.block_pbin, self.block_val, self.block_val_error, self.edge_index,
                   self.edges)
        # get edge_index
        self.edge_index = np.array([np.where(self.time >= self.edges[i])[0][0] 
                                    for i,_ in enumerate(self.edges)])
        #change last entry such that loop over [j:j+1] gives all blocks
        self.edge_index[-1] += 1

        # determine flux value (mean) and error (Gaussian propagation) for each block
        self.block_val = np.zeros(len(self.edge_index)-1)  
        self.block_val_error = np.zeros(len(self.edge_index)-1) 
        for j in range(len(self.edge_index)-1):
            self.block_val[j] = np.mean(self.flux[self.edge_index[j]: self.edge_index[j+1]])
            self.block_val_error[j] = (np.sqrt(np.sum(self.flux_error[self.edge_index[j]:
                                                     self.edge_index[j+1]]**2))
                                      / (self.edge_index[j+1]-self.edge_index[j]))

        # create block-per-bin array corresponding to flux
        self.block_pbin = np.zeros(len(self.flux))
        for k,_ in enumerate(self.block_val):
            self.block_pbin[self.edge_index[k] : self.edge_index[k+1]] = self.block_val[k]
        logging.debug('got block parameters for light curve')

        return(self.block_pbin, self.block_val, self.block_val_error, self.edge_index, self.edges)

    #----------------------------------------------------------------------------------------------
    def get_bblocks_above(self, threshold, pass_gamma_value=None, pass_p0_value=None):
        """
        ATTENTION! This returns artificial flux values! Use cautiously if at all..
        Note: get_bblocks has to be applied first
        Determine Bayesian blocks for light curve but set all blocks that are lower than threshold
        to that threshold (i.e. set small block_val to threshold and neglect edges under threshold)
        -> leaves only significant variations wrt threshold = flares?
        """ 
        self.block_pbin = np.where(self.block_pbin > threshold, self.block_pbin, threshold)
        self.block_val = np.where(self.block_val > threshold, self.block_val, threshold)
        block_mask = np.ones(len(self.block_val), dtype = bool)
        edge_mask = np.ones(len(self.edges), dtype=bool)
        for i in range(len(self.block_val)-1):
            if self.block_val[i] == threshold and self.block_val[i+1] == threshold:
                block_mask[i+1] = False
                edge_mask[i+1] = False
        self.block_val = self.block_val[block_mask]
        self.block_val_error = self.block_val_error[block_mask]
        self.edge_index = self.edge_index[edge_mask]
        self.edges = self.edges[edge_mask]
        return(self.block_pbin, self.block_val, self.block_val_error, self.edge_index, self.edges)

    #----------------------------------------------------------------------------------------------
    def plot_bblocks(self, bb_color='steelblue', data_color='k', data_label='obs flux',
                     size=1, ax=None):
        if ax is None:
                ax = plt.gca()
        try:   
            ax.step(self.time, self.block_pbin, where='mid', linewidth=1*size, label='bblocks', 
            	     color=bb_color, zorder=1000)
            ax.errorbar(x=self.time, y=self.flux, yerr=self.flux_error, label=data_label, 
            	         ecolor=data_color, elinewidth=1*size, linewidth=0, marker='+', 
                         markersize=3*size, color=data_color)
        except AttributeError:
            raise AttributeError('Initialize Bayesian blocks with .get_bblocks() first!')

    #----------------------------------------------------------------------------------------------
    def bb_i(self, t):
        """
        Convert time to index of corresponding Bayesian block (e.g. block_value of peak_time)
        use bb_i_start/bb_i_end to make sure you get the block left/right outside of hop
        this works fine for flip, halfclap, and sharp but *NOT for BASELINE* (-> block inside hop)
        """
        if t == self.edges[0]:
            return(int(0))
        else:
            block_index = [
                e for e in range(len(self.edges)-1) if t > self.edges[e] and t <= self.edges[e+1]]
            return(int(block_index[0]))

    def bb_i_start(self,t):
        """
        if time = edge -> take block on the left
        ATTENTION: for baseline method this is first block of hop -> use bb_i() instead (works)
        """
        block_index = [
            e for e in range(len(self.edges)-1) if t >= self.edges[e] and t < self.edges[e+1]]
        return(int(block_index[0]))

    def bb_i_end(self,t):
        """
        if time = edge -> take block on the right
        ATTENTION: for baseline method this is last block of hop - use bb_i() instead (TBD)
        """
        block_index = [
            e for e in range(len(self.edges)-1) if t > self.edges[e] and t <= self.edges[e+1]]
        return(int(block_index[0]))
    
    #----------------------------------------------------------------------------------------------
    def find_hop(self, method='half', lc_edges='neglect', baseline=None):
        if method == 'baseline':
            if baseline is None:
                self.baseline = np.mean(self.flux)
            else:
                self.baseline = baseline
            hopfinder = HopFinderBaseline(lc_edges)
        if method == 'half':
            hopfinder = HopFinderHalf(lc_edges)
        if method == 'sharp':
            hopfinder = HopFinderSharp(lc_edges)
        if method == 'flip':
            hopfinder = HopFinderFlip(lc_edges)
        self.hops = hopfinder.find(self)
        return self.hops

    #----------------------------------------------------------------------------------------------
    def plot_hop(self, ax=None, **kwargs):
        """
        Plot shaded area for all hops in light curve
        """
        if self.hops is None:
            return # no hop in this lc
        if ax is None:
            ax = plt.gca()
        for i,hop in enumerate(self.hops):
            x = np.linspace(hop.start_time, hop.end_time)
            y = np.ones(len(x)) * np.max(self.flux)
            y1 = np.min(self.flux)
            if i == 0:
                ax.fill_between(x, y, y1, step="mid", color='lightsalmon', alpha=0.2,
                                 label='hop', zorder=0)
            if i == 1:
                ax.fill_between(x, y, y1, step="mid", color='orchid', alpha=0.2, label='hop',
                                 zorder=0)
            elif i % 2:
                ax.fill_between(x, y, y1, step="mid", color='orchid', alpha=0.2, zorder=0)
            elif i != 0:
                ax.fill_between(x, y, y1, step="mid", color='lightsalmon', alpha=0.2, zorder=0)
        #ax.set_title(lc.name, hop.method)

    #----------------------------------------------------------------------------------------------
    def plot_all_hop(self, gamma_value=None, p0_value=0.05, lc_edges='neglect'):
        """
        Plot all HOP methods in one figure for comparison
        """
        fig = plt.figure(0,(15,9))
        plt.suptitle('All HOP methods', fontsize=16)

        ax0 = fig.add_subplot(511)
        self.find_hop('baseline')
        self.plot_bblocks()
        self.plot_hop()
        plt.ylabel('baseline')

        ax1 = fig.add_subplot(512)
        self.find_hop('half')
        self.plot_bblocks()
        self.plot_hop()
        plt.ylabel('half')

        ax2 = fig.add_subplot(513)
        self.find_hop('flip')
        self.plot_bblocks()
        self.plot_hop()
        plt.ylabel('flip')

        ax3 = fig.add_subplot(514)
        self.find_hop('sharp')
        self.plot_bblocks()
        self.plot_hop()
        plt.ylabel('sharp')
        fig.subplots_adjust(hspace=0)





import numpy as np
from lmfit import Model #https://lmfit.github.io/lmfit-py/
from matplotlib import pyplot as plt

class Hopject: 
    '''
    HOP Class
    ---------
    Segment in a light curve, i.e. a group of Bayesian blocks, that represent a flare ( HOP group)
    For definition of start, peak and end of the flare check out the LightCurve class
    '''
    def __init__(self, hop_params, lc, method=None): #e.g. Hopject(lc.get_hop_method[0], lc)
        self.start_time, self.peak_time, self.end_time = hop_params        
        self.lc = lc
        self.z = lc.z
        self.name = lc.name
        self.telescope = lc.telescope
        self.method = method #e.g. half or flip
        self.iis = np.where(np.logical_and(lc.time > self.start_time, 
        								   lc.time < self.end_time))
        self.time = lc.time[self.iis]
        self.flux = lc.flux[self.iis]
        self.flux_error = lc.flux_error[self.iis]
        self.n_bins = len(self.time)
        self.coverage = self.n_bins / (self.end_time - self.start_time)
        self.n_blocks = lc.bb_i_end(self.end_time) - lc.bb_i_start(self.start_time) 
        # e.g. one-block hop: 5 - 3 = 2 

        self.dur = self.end_time - self.start_time
        self.rise_time = self.peak_time - self.start_time
        self.decay_time = self. end_time - self. peak_time
        self.asym = (self.rise_time - self.decay_time)/(self.rise_time + self.decay_time)

        self.start_flux = lc.block_val[lc.bb_i_start(self.start_time)]
        self.peak_flux = lc.block_val[lc.bb_i(self.peak_time)]
        self.end_flux = lc.block_val[lc.bb_i_end(self.end_time)]
        self.rise_flux = self.peak_flux - self.start_flux
        self.decay_flux = self.peak_flux - self.end_flux

    #----------------------------------------------------------------------------------------------
    def plot_hop(self, ax=None):
        """
        Plot the snip of light curve with this flare
        """
        if ax is None:
            ax = plt.gca()
        self.lc.plot_bblocks(ax=ax)
        ax.set_xlim(self.start_time - self.dur/2, self.end_time + self.dur/2)
        x = np.linspace(self.start_time, self.end_time)
        y = np.ones(len(x)) * self.peak_flux
        y1 = np.zeros(len(x))
        ax.fill_between(x, y, y1, step="mid", color='lightsalmon', alpha=0.2, label='hop', zorder=0)

    #----------------------------------------------------------------------------------------------
    def exp_rd(self, t, amp, t_0, t_r, t_d):
        """
        exponential rise and exponential decay
        with flux = amp/2 at t=t_0 (peak could be shifted due to ratio of t_r and t_d)
        $F(x) = amp * ( exp(\frac{t-cen}{t_{decay}}) + exp(\frac{-t+cen}{t_{rise}}) )^{-1}$
        """
        return(amp * ( np.exp((t-t_0)/t_d) + np.exp((-t+t_0)/t_r))**-1)

    def get_exp_fit(self):
        """exponential fit on data points inside hop"""
        exp_model = Model(self.exp_rd)
        params = exp_model.make_params(amp=self.peak_flux, 
                                       t_0=self.peak_time, 
                                       t_r=self.rise_time,
                                       t_d=self.decay_time)
        result = exp_model.fit(self.flux, params, t=self.time, weights=1/self.flux_error)
        self.exp_tr = result.params['t_r'].value
        self.exp_td = result.params['t_d'].value
        self.exp_amp = result.params['amp'].value
        self.exp_t0 = result.params['t_0'].value
        self.exp_chisqr = result.chisqr # $\chi^2 = \sum_i \frac{(O-C)^2}{\sigma_i^2}$
        self.exp_redchi = result.redchi # $\chi^2_R = \sum_i \frac{(O-C)^2}{\sigma_i^2}$ 

        return result

    def plot_exp_fit(self, plotpoints=200, ax=None, **plot_kwargs):
        self.get_exp_fit()
        x_plot = np.linspace(self.start_time, self.end_time, plotpoints)
        y_plot = self.exp_rd(t=x_plot, amp=self.exp_amp, t_0=self.exp_t0, 
                             t_r=self.exp_tr, t_d=self.exp_td)
        if ax is None:
            ax = plt.gca()
        ax.plot(x_plot,y_plot, marker='', zorder=13148, **plot_kwargs)

    #----------------------------------------------------------------------------------------------
    def gaussian(self, x, amp, cen, wid):
        """1-d gaussian: gaussian(x=data, amp, cen, wid)"""
        return (amp / (np.sqrt(2*np.pi) * wid)) * np.exp(-(x-cen)**2 / (2*wid**2))

    def get_gauss_fit(self):
        """gaussian fit on data points inside hop"""
        x = self.time
        y = self.flux
        y_error = self.flux_error
        gmodel = Model(self.gaussian)
        params = gmodel.make_params(cen=np.mean(x), 
                                    amp=self.peak_flux, 
                                    wid=np.std(x))
        result = gmodel.fit(y, params, x=x, weights=y_error)#Poisson error: 1/np.where(y == 0, 1, y))
        self.gauss_amp = result.params['amp'].value
        self.gauss_mu = result.params['cen'].value  
        self.gauss_sigma = result.params['wid'].value
        self.gauss_chisqr = result.chisqr
        self.gauss_redchi = result.redchi
        return(result)

    def plot_gauss_fit(self, plotpoints=200, ax=None, **plot_kwargs):
        self.get_gauss_fit()
        x_plot = np.linspace(self.start_time, self.end_time, plotpoints)
        y_plot = self.gaussian(x_plot, self.gauss_amp, self.gauss_mu, self.gauss_sigma)
        if ax is None:
            ax = plt.gca()
        ax.plot(x_plot,y_plot, marker='', zorder=13148, **plot_kwargs)
        return()




import numpy as np
from matplotlib import pyplot as plt
import astropy.visualization.hist as fancy_hist
#https://docs.astropy.org/en/stable/api/astropy.visualization.hist.html
from lightcurves.LC import LightCurve
from lightcurves.HOP import Hopject
import logging
logging.basicConfig(level=logging.ERROR) #see LC.py


class LC_Set:
    '''
    Light Curve Set
    ===============
    Process and analyze all Hopjects (see HOP.py) in a set of Light Curves (see LC_new.py)
    Determine the distribution of all HOP parameters in the LC_Set
    Plot the resulting distributions of HOP properties in a fancy way

    lcs:
        list or np.array of Light Curve Objects
        Note: get_bblocks() needs to be applied to all lcs first! 
        For example:
            lcs = np.zeros(len(files), dtype = object)
            for i in files:
                lc = LightCurve(time[i], flux[i], flux_error[i])
                lc.get_bblocks()
                lcs[i] = lc

    hop_method:
        a) 'baseline'
            Determine start_time/end_time to be where flux exceeds/goes under baseline 
        b) 'half'
            Determine start/end of flare to be at center of valley block
        c) 'flip'
            Extrapolate behavior of flare by flipping adjacent block onto valley block
            Note: half method is used to avoid overlap (i.e. when flip > 1/2 valley block)
        d) 'sharp'
            Neglect valley block

    lc_edges:
        a) neglect:
            single start and end times are neglected
            peaks without start or end time are neglected
        b) add:
            single start and end times are neglected
            peaks without start/end: start/end is added in beginning/end of light curve

    baseline:
        e.g. mean of flux (default), median of flux, quiescent background ...

    block_min: 
        Minimal number of blocks to be a flare, e.g. block_min = 2 -> no single-block flares
    '''
    def __init__(self, lcs, hop_method='flip', lc_edges='neglect', baseline='mean', block_min=1):
        # lc.get_bblock needs to be run already for each lc (do that in initialization)!!
        # check weather get_bblockswas run for lc in lcs; return error if not
        self.lcs = lcs
        mom_lc = []
        hopjects = []

        #make sure Bayesian blocks have been initialized before
        for i,lc in enumerate(lcs):
            try: 
                lc.block_pbin
            except AttributeError:
                raise AttributeError('Initialize Bayesian blocks for all LCs first!')

            logging.debug(str(i))
            #optional: multiprocessing for hoparound of each lc here
            if hop_method == 'baseline' and baseline == 'mean':
                hops = lc.get_hop_baseline(np.mean(lc.flux),lc_edges)
            elif hop_method == 'baseline':
                hops = lc.get_hop_bl(baseline,lc_edges)
            elif hop_method == 'half':
                hops = lc.get_hop_half(lc_edges)
            elif hop_method == 'flip':
                hops = lc.get_hop_flip(lc_edges)
            elif hop_method == 'sharp':
                hops = lc.get_hop_sharp(lc_edges)
            if hops is None:
                logging.info(str(i)+ ' no hop found; not variable enough')
                continue #skip this lc
            for hop_params in hops:
                hop = Hopject(hop_params, lc) #hop = hop_params (start, peak, end)
                hop.get_exp_fit()
                hopjects.append(hop)
                mom_lc.append(i)
        self.n_blocks = np.array([h.n_blocks for h in hopjects])
        mask = np.where(self.n_blocks > block_min)
        # eg one-block hop: n_blocks = end_block - start_block 
        #                            = 5 - 3 = 2 !> 2 (minimum blocks of hop)

        # Patrick: lieber erst implementieren, wenns gebraucht wird bzw mit property decorator siehe unten
        self.mom_lc = np.array(mom_lc)[mask]
        self.hopjects = np.array(hopjects, dtype = object)[mask]
        self.dur = np.array([h.dur for h in hopjects])[mask]
        self.rise_time = np.array([h.rise_time for h in hopjects])[mask]
        self.decay_time = np.array([h.decay_time for h in hopjects])[mask]
        self.asym = np.array([h.asym for h in hopjects])[mask]
        self.start_flux = np.array([h.start_flux for h in hopjects])[mask]
        self.peak_flux = np.array([h.peak_flux for h in hopjects])[mask]
        self.end_flux = np.array([h.end_flux for h in hopjects])[mask]
        self.rise_flux = np.array([h.rise_flux for h in hopjects])[mask]
        self.decay_flux = np.array([h.decay_flux for h in hopjects])[mask]
        self.z = np.array([h.z for h in hopjects])[mask]

        #self.exp_tr = np.array([h.exp_tr for h in hopjects if not h.exp_tr is None])[mask]
        self.exp_td = np.array([h.exp_td for h in hopjects])[mask]
        self.exp_amp = np.array([h.exp_amp for h in hopjects])[mask]
        self.exp_t0 = np.array([h.exp_t0 for h in hopjects])[mask]
        self.exp_chisqr = np.array([h.exp_chisqr for h in hopjects])[mask]
        self.exp_redchi = np.array([h.exp_redchi for h in hopjects])[mask]

    @property
    def exp_tr(self):
        return np.array([h.exp_tr for h in hopjects])[mask]
    
    #----------------------------------------------------------------------------------------------
    def zcor(self, times): #times = e.g. LC_Set.dur
        if len(np.where(np.isnan(self.z) == True)[0]) > 0:
            print('Error: not all LCs have a redshift')
        else:
            times_intr = times / (1 + self.z)
            return(times_intr)

    #----------------------------------------------------------------------------------------------
    def plot_asym(self, N_bins=None, dens=True):
        histo, fancy_bins, p = fancy_hist(self.asym, bins='blocks', density=dens, histtype='step',
                                          label='Bayesian binning')
        if N_bins:
            plt.hist(self.asym, N_bins)
        else:
            histo, fancy_bins, p = fancy_hist(self.asym, bins='knuth', density=dens, edgecolor='k',
                                              color='hotpink', label='Knuth bins')

    def plot_dur(self, N_bins=None, dens=True):
        histo, fancy_bins, p = fancy_hist(self.dur, bins='blocks', density=dens, histtype='step')
        if N_bins:
            plt.hist(self.asym, N_bins)
        else:
            histo, fancy_bins, p = fancy_hist(self.dur, bins='knuth', density=dens, edgecolor='k',
                                              color='limegreen')

